from collections import OrderedDict

import astropy.coordinates as coord
import astropy.units as u
import matplotlib.pyplot as plt

# import mpl_toolkits.basemap as bm
import numpy as np
import spherical_geometry.polygon as sp
from astropy.table import Table
import astropy.time as time

from .gbm_detector import BGO0, BGO1
from .gbm_detector import NaI0, NaI1, NaI2, NaI3, NaI4, NaI5
from .gbm_detector import NaI6, NaI7, NaI8, NaI9, NaIA, NaIB
from .gbm_frame import GBMFrame


from gbmgeometry.utils.gbm_time import GBMTime

#import seaborn as sns

_det_color_cycle = np.linspace(0, 1, 12)


class GBM(object):
    def __init__(self, quaternion, sc_pos=None, gbm_time=None):
        """
        

        :param quaternion: 
        :param sc_pos: 
        :param gbm_time: 
        :returns: 
        :rtype: 

        """

        if gbm_time is not None:

            if isinstance(gbm_time, str):

                self._gbm_time = GBMTime.from_UTC_fits(gbm_time)

            else:

                # assuming MET

                self._gbm_time = GBMTime.from_MET(gbm_time)

        else:

            self._gbm_time = None

        if self._gbm_time is not None:

            self.n0 = NaI0(quaternion, sc_pos, self._gbm_time.time)
            self.n1 = NaI1(quaternion, sc_pos, self._gbm_time.time)
            self.n2 = NaI2(quaternion, sc_pos, self._gbm_time.time)
            self.n3 = NaI3(quaternion, sc_pos, self._gbm_time.time)
            self.n4 = NaI4(quaternion, sc_pos, self._gbm_time.time)
            self.n5 = NaI5(quaternion, sc_pos, self._gbm_time.time)
            self.n6 = NaI6(quaternion, sc_pos, self._gbm_time.time)
            self.n7 = NaI7(quaternion, sc_pos, self._gbm_time.time)
            self.n8 = NaI8(quaternion, sc_pos, self._gbm_time.time)
            self.n9 = NaI9(quaternion, sc_pos, self._gbm_time.time)
            self.na = NaIA(quaternion, sc_pos, self._gbm_time.time)
            self.nb = NaIB(quaternion, sc_pos, self._gbm_time.time)
            self.b0 = BGO0(quaternion, sc_pos, self._gbm_time.time)
            self.b1 = BGO1(quaternion, sc_pos, self._gbm_time.time)

        else:

            self.n0 = NaI0(quaternion, sc_pos, None)
            self.n1 = NaI1(quaternion, sc_pos, None)
            self.n2 = NaI2(quaternion, sc_pos, None)
            self.n3 = NaI3(quaternion, sc_pos, None)
            self.n4 = NaI4(quaternion, sc_pos, None)
            self.n5 = NaI5(quaternion, sc_pos, None)
            self.n6 = NaI6(quaternion, sc_pos, None)
            self.n7 = NaI7(quaternion, sc_pos, None)
            self.n8 = NaI8(quaternion, sc_pos, None)
            self.n9 = NaI9(quaternion, sc_pos, None)
            self.na = NaIA(quaternion, sc_pos, None)
            self.nb = NaIB(quaternion, sc_pos, None)
            self.b0 = BGO0(quaternion, sc_pos, None)
            self.b1 = BGO1(quaternion, sc_pos, None)

        self._detectors = OrderedDict(
            n0=self.n0,
            n1=self.n1,
            n2=self.n2,
            n3=self.n3,
            n4=self.n4,
            n5=self.n5,
            n6=self.n6,
            n7=self.n7,
            n8=self.n8,
            n9=self.n9,
            na=self.na,
            nb=self.nb,
            b0=self.b0,
            b1=self.b1,
        )
        self._quaternion = quaternion
        self._sc_pos = sc_pos

    def set_quaternion(self, quaternion):
        """FIXME! briefly describe function

        :param quaternion: 
        :returns: 
        :rtype: 

        """

        for key in self._detectors.keys():
            self._detectors[key].set_quaternion(quaternion)

        self._quaternion = quaternion

    def set_sc_pos(self, sc_pos):
        """FIXME! briefly describe function

        :param sc_pos: 
        :returns: 
        :rtype: 

        """
      

        for key in self._detectors.keys():
            self._detectors[key].set_sc_pos(sc_pos)

        self._sc_pos = sc_pos

    def get_good_detectors(self, point, fov):
        """
        Returns a list of detectors containing the point in the FOV

        Parameters
        ----------
        point
        fov

        Returns
        -------

        """

        good_detectors = self._contains_point(point, fov)

        return good_detectors

    def get_fov(self, radius, fermi_frame=False):
        """
        Parameters
        ----------
        fermi_frame
        radius

        """

        polys = []

        for key in self._detectors.keys():

            if key[0] == "b":
                this_rad = 90

            else:
                this_rad = radius

            polys.append(self._detectors[key].get_fov(this_rad, fermi_frame))

        polys = np.array(polys)

        return polys

    def get_good_fov(self, point, radius, fermi_frame=False):
        """
        Returns the detectors that contain the given point
        for the given angular radius

        Parameters
        ----------
        point
        radius


        """

        good_detectors = self._contains_point(point, radius)

        polys = []

        for key in good_detectors:
            polys.append(self._detectors[key].get_fov(radius, fermi_frame))

        return [polys, good_detectors]

    def get_sun_angle(self, keys=None):

        """

        Returns
        -------

        """
        angles = []

        if keys is None:

            for key in self._detectors.keys():
                angles.append(self._detectors[key].sun_angle)

        else:

            for key in keys:
                angles.append(self._detectors[key].sun_angle)

        return angles

    def get_centers(self, keys=None):

        """

        Returns
        -------

        """
        centers = []

        if keys is None:

            for key in self._detectors.keys():
                centers.append(self._detectors[key].get_center())

        else:

            for key in keys:
                centers.append(self._detectors[key].get_center())

        return centers

    def get_separation(self, source):
        """
        Get the andular separation of the detectors from a point
        Parameters
        ----------
        source

        Returns
        -------

        """

        tab = Table(names=["Detector", "Separation"], dtype=["|S2", np.float64])

        for key in self._detectors.keys():
            sep = self._detectors[key].get_center().separation(source)
            tab.add_row([key, sep])

        tab["Separation"].unit = u.degree

        tab.sort("Separation")

        return tab

    def get_earth_points(self, fermi_frame=False):
        """

        Returns
        -------

        """

        if self._sc_pos is not None:

            self._calc_earth_points(fermi_frame)

            return self._earth_points

        else:
            print("No spacecraft position set")

    def _calc_earth_points(self, fermi_frame):

        xyz_position = coord.SkyCoord(
            x=self._sc_pos[0],
            y=self._sc_pos[1],
            z=self._sc_pos[2],
            frame="icrs",
            representation="cartesian",
        )

        earth_radius = 6371.0 * u.km
        fermi_radius = np.sqrt((self._sc_pos ** 2).sum())

        horizon_angle = 90 - np.rad2deg(
            np.arccos((earth_radius / fermi_radius).to(u.dimensionless_unscaled)).value
        )

        horizon_angle = (180 - horizon_angle) * u.degree

        num_points = 300

        ra_grid_tmp = np.linspace(0, 360, num_points)

        dec_range = [-90, 90]
        cosdec_min = np.cos(np.deg2rad(90.0 + dec_range[0]))
        cosdec_max = np.cos(np.deg2rad(90.0 + dec_range[1]))

        v = np.linspace(cosdec_min, cosdec_max, num_points)
        v = np.arccos(v)
        v = np.rad2deg(v)
        v -= 90.0

        dec_grid_tmp = v

        ra_grid = np.zeros(num_points ** 2)
        dec_grid = np.zeros(num_points ** 2)

        itr = 0
        for ra in ra_grid_tmp:
            for dec in dec_grid_tmp:
                ra_grid[itr] = ra
                dec_grid[itr] = dec
                itr += 1

        if fermi_frame:
            all_sky = coord.SkyCoord(
                Az=ra_grid,
                Zen=dec_grid,
                frame=GBMFrame(quaternion=self._quaternion),
                unit="deg",
            )
        else:
            all_sky = coord.SkyCoord(ra=ra_grid, dec=dec_grid, frame="icrs", unit="deg")

        condition = all_sky.separation(xyz_position) > horizon_angle

        # self.seps = all_sky.separation(xyz_position)

        self._earth_points = all_sky[condition]

    @property
    def detectors(self):

        return self._detectors

    def _contains_point(self, point, radius):
        """
        returns detectors that contain a points
        """

        condition = []

        steps = 500

        for key in self._detectors.keys():

            if key[0] == "b":
                this_rad = 90
            else:
                this_rad = radius

            j2000 = self._detectors[key]._center.icrs

            poly = sp.SphericalPolygon.from_cone(
                j2000.ra.value, j2000.dec.value, this_rad, steps=steps
            )

            if poly.contains_point(point.cartesian.xyz.value):
                condition.append(key)

        return condition


# def get_legal_pairs():
#     """
#     Plots the legal pairs of detectors for GBM observations

#     Returns
#     -------

#     """

#     dlp = np.array(
#         [
#             [0, 274, 39, 171, 12, 29, 0, 5, 1, 6, 1, 0],
#             [258, 0, 233, 55, 4, 100, 2, 1, 1, 12, 27, 0],
#             [55, 437, 0, 2, 2, 311, 0, 1, 1, 13, 235, 0],
#             [215, 80, 3, 0, 330, 107, 4, 8, 19, 2, 1, 0],
#             [13, 4, 8, 508, 0, 269, 2, 29, 236, 0, 1, 0],
#             [44, 188, 337, 166, 279, 0, 0, 0, 0, 0, 0, 0],
#             [0, 1, 1, 2, 2, 0, 0, 238, 46, 180, 12, 33],
#             [0, 2, 0, 18, 35, 0, 222, 0, 221, 61, 3, 109],
#             [0, 0, 1, 16, 215, 0, 51, 399, 0, 4, 2, 303],
#             [3, 18, 21, 4, 0, 0, 190, 82, 1, 0, 324, 110],
#             [1, 25, 191, 0, 0, 0, 16, 6, 4, 516, 0, 293],
#             [0, 0, 0, 0, 0, 0, 32, 147, 297, 138, 263, 0],
#         ]
#     )

#     sns.heatmap(dlp, annot=True, fmt="d", cmap="YlGnBu")
#     plt.ylabel("NaI")
#     plt.xlabel("NaI")



import astropy.coordinates as coord
import astropy.units as u
import numpy as np
from astropy.coordinates import BaseCoordinateFrame, Attribute, RepresentationMapping
from astropy.coordinates import frame_transform_graph


class GBMFrame(BaseCoordinateFrame):
    """
    
    Fermi GBM Frame

    Parameters
    ----------
    representation : `BaseRepresentation` or None
        A representation object or None to have no data (or use the other keywords)
  
    """

    default_representation = coord.SphericalRepresentation

    frame_specific_representation_info = {
        "spherical": [
            RepresentationMapping(
                reprname="lon", framename="lon", defaultunit=u.degree
            ),
            RepresentationMapping(
                reprname="lat", framename="lat", defaultunit=u.degree
            ),
            RepresentationMapping(
                reprname="distance", framename="DIST", defaultunit=None
            ),
        ],
        "unitspherical": [
            RepresentationMapping(
                reprname="lon", framename="lon", defaultunit=u.degree
            ),
            RepresentationMapping(
                reprname="lat", framename="lat", defaultunit=u.degree
            ),
        ],
        "cartesian": [
            RepresentationMapping(reprname="x", framename="SCX"),
            RepresentationMapping(reprname="y", framename="SCY"),
            RepresentationMapping(reprname="z", framename="SCZ"),
        ],
    }

    # Specify frame attributes required to fully specify the frame
    sc_pos_X = Attribute(default=None)
    sc_pos_Y = Attribute(default=None)
    sc_pos_Z = Attribute(default=None)

    quaternion_1 = Attribute(default=None)
    quaternion_2 = Attribute(default=None)
    quaternion_3 = Attribute(default=None)
    quaternion_4 = Attribute(default=None)

    # equinox = TimeFrameAttribute(default='J2000')

    @staticmethod
    def _set_quaternion(q1, q2, q3, q4):
        sc_matrix = np.zeros((3, 3))

        sc_matrix[0, 0] = q1 ** 2 - q2 ** 2 - q3 ** 2 + q4 ** 2
        sc_matrix[0, 1] = 2.0 * (q1 * q2 + q4 * q3)
        sc_matrix[0, 2] = 2.0 * (q1 * q3 - q4 * q2)
        sc_matrix[1, 0] = 2.0 * (q1 * q2 - q4 * q3)
        sc_matrix[1, 1] = -(q1 ** 2) + q2 ** 2 - q3 ** 2 + q4 ** 2
        sc_matrix[1, 2] = 2.0 * (q2 * q3 + q4 * q1)
        sc_matrix[2, 0] = 2.0 * (q1 * q3 + q4 * q2)
        sc_matrix[2, 1] = 2.0 * (q2 * q3 - q4 * q1)
        sc_matrix[2, 2] = -(q1 ** 2) - q2 ** 2 + q3 ** 2 + q4 ** 2

        return sc_matrix


@frame_transform_graph.transform(coord.FunctionTransform, GBMFrame, coord.ICRS)
def gbm_to_j2000(gbm_coord, j2000_frame):
    """ Compute the transformation from heliocentric Sgr coordinates to
        spherical Galactic.
    """

    sc_matrix = gbm_coord._set_quaternion(
        gbm_coord.quaternion_1,
        gbm_coord.quaternion_2,
        gbm_coord.quaternion_3,
        gbm_coord.quaternion_4,
    )

    # X,Y,Z = gbm_coord.cartesian

    pos = gbm_coord.cartesian.xyz.value

    X0 = np.dot(sc_matrix[:, 0], pos)
    X1 = np.dot(sc_matrix[:, 1], pos)
    X2 = np.clip(np.dot(sc_matrix[:, 2], pos), -1.0, 1.0)

    # dec = np.arcsin(X2)

    dec = np.pi / 2.0 - np.arccos(X2)

    idx = np.logical_and(np.abs(X0) < 1e-6, np.abs(X1) < 1e-6)

    ra = np.zeros_like(dec)

    ra[~idx] = np.arctan2(X1, X0) % (2 * np.pi)

    return coord.ICRS(ra=ra * u.radian, dec=dec * u.radian)


@frame_transform_graph.transform(coord.FunctionTransform, coord.ICRS, GBMFrame)
def j2000_to_gbm(j2000_frame, gbm_coord):
    """ Compute the transformation from heliocentric Sgr coordinates to
        spherical Galactic.
    """

    sc_matrix = gbm_coord._set_quaternion(
        gbm_coord.quaternion_1,
        gbm_coord.quaternion_2,
        gbm_coord.quaternion_3,
        gbm_coord.quaternion_4,
    )

    pos = j2000_frame.cartesian.xyz.value

    X0 = np.dot(sc_matrix[0, :], pos)
    X1 = np.dot(sc_matrix[1, :], pos)
    X2 = np.clip(np.dot(sc_matrix[2, :], pos), -1.0, 1.0)
    el = np.pi / 2.0 - np.arccos(X2)  # convert to proper frame

    idx = np.logical_and(np.abs(X0) < 1e-6, np.abs(X1) < 1e-6)

    az = np.zeros_like(el)

    az[~idx] = np.arctan2(X1, X0) % (2 * np.pi)

    az[np.rad2deg(el) == 90.0] = 0.0

    return GBMFrame(
        lon=az * u.radian,
        lat=el * u.radian,
        quaternion_1=gbm_coord.quaternion_1,
        quaternion_2=gbm_coord.quaternion_2,
        quaternion_3=gbm_coord.quaternion_3,
        quaternion_4=gbm_coord.quaternion_4,
    )



__author__ = "drjfunk"
import numpy as np
from astropy.coordinates import (
    SkyCoord,
    get_sun,
    get_body,
    get_body_barycentric,
    get_moon,
)
from spherical_geometry.polygon import SphericalPolygon
import astropy.units as u

from .gbm_frame import GBMFrame


class GBMDetector(object):
    def __init__(self, name, quaternion, sc_pos=None, time=None):
        """
        
        :param name: 
        :param quaternion: 
        :param sc_pos: 
        :param time: 
        """

        self._name = name

        self.update_position(quaternion, sc_pos, time)

        self._xyz = np.array(
            [
                np.cos(self._az) * np.sin(self._zen),
                np.sin(self._az) * np.sin(self._zen),
                np.cos(self._zen),
            ]
        )

    def update_position(self, quaternion, sc_pos=None, time=None):
        """
        
        :param quaternion: 
        :param sc_pos: 
        :param time: 
        :return: 
        """

        self._time = time

        q1, q2, q3, q4 = quaternion

        if sc_pos is not None:
            scx, scy, scz = sc_pos
            self._heigth = np.sqrt(scx * scx + scy * scy + scz * scz)

        else:
            scx = None
            scy = None
            scz = None
        self._quaternion = quaternion
        self._sc_pos = sc_pos
        # define the direction of the detector in the GBMFrame
        self._center = SkyCoord(
            lon=self._az * u.deg,
            lat=self._zen * u.deg,
            unit="deg",
            frame=GBMFrame(
                quaternion_1=q1,
                quaternion_2=q2,
                quaternion_3=q3,
                quaternion_4=q4,
                sc_pos_X=scx,
                sc_pos_Y=scy,
                sc_pos_Z=scz,
            ),
        )
        if self._time is not None:
            # we can calculate the sun position
            # in GCRS
            tmp_sun = get_sun(self._time)
            # in ICRS
            self._sun_position_icrs = SkyCoord(
                tmp_sun.ra.deg,
                tmp_sun.dec.deg,
                unit="deg",
                frame="gcrs",
                obstime=self._time,
            ).icrs
            # in CenterFrame
            self._sun_position = self._sun_position_icrs.transform_to(
                self._center.frame
            )

            # position of earth in satellite frame:

        if sc_pos is not None:
            

            self._earth_pos_norm = self.geo_to_gbm(
                -self._sc_pos / np.linalg.norm(self._sc_pos)
            )
            scxn, scyn, sczn = self._earth_pos_norm
            earth_theta = np.arccos(sczn / np.sqrt(scxn * scxn + scyn * scyn + sczn * sczn))
            earth_phi = np.arctan2(scyn, scxn)
            earth_ra = np.rad2deg(earth_phi)
            if earth_ra < 0:
                earth_ra = earth_ra + 360
            earth_dec = 90 - np.rad2deg(earth_theta)

            # earth as SkyCoord
            self._earth_position = SkyCoord(
                lon=earth_ra * u.deg,
                lat=earth_dec * u.deg,
                unit="deg",
                frame=GBMFrame(
                    quaternion_1=q1,
                    quaternion_2=q2,
                    quaternion_3=q3,
                    quaternion_4=q4,
                    sc_pos_X=scx,
                    sc_pos_Y=scy,
                    sc_pos_Z=scz,
                ),
            )
        self._scx = scx
        self._scy = scy
        self._scz = scz

    def set_quaternion(self, quaternion):
        """
        Parameters
        ----------
        quaternion

        """

        self._quaternion = quaternion

        q1, q2, q3, q4 = quaternion

        if self._sc_pos is not None:
            scx, scy, scz = self._sc_pos

        else:
            scx = None
            scy = None
            scz = None

        self._center = SkyCoord(
            lon=self._az * u.deg,
            lat=self._zen * u.deg,
            unit="deg",
            frame=GBMFrame(
                quaternion_1=q1,
                quaternion_2=q2,
                quaternion_3=q3,
                quaternion_4=q4,
                sc_pos_X=scx,
                sc_pos_Y=scy,
                sc_pos_Z=scz,
            ),
        )

    def set_sc_pos(self, sc_pos):
        """
        Parameters
        ----------
        quaternion

        """

        q1, q2, q3, q4 = self._quaternion

        if sc_pos is not None:
            scx, scy, scz = sc_pos

        else:
            scx = None
            scy = None
            scz = None

        self._center = SkyCoord(
            lon=self._az * u.deg,
            lat=self._zen * u.deg,
            unit="deg",
            frame=GBMFrame(
                quaternion_1=q1,
                quaternion_2=q2,
                quaternion_3=q3,
                quaternion_4=q4,
                sc_pos_X=scx,
                sc_pos_Y=scy,
                sc_pos_Z=scz,
            ),
        )

    def get_fov(self, radius, fermi_frame=False):
        """
        Returns
        -------
        array of RA and DEC

        """

        steps = 500

        if fermi_frame:
            fermi = self._center

            poly = SphericalPolygon.from_cone(
                fermi.lon.value, fermi.lat.value, radius, steps=steps
            )

        else:

            j2000 = self._center.icrs

            poly = SphericalPolygon.from_cone(
                j2000.ra.value, j2000.dec.value, radius, steps=steps
            )

        # ra, dec
        return [p for p in poly.to_radec()][0]

    def get_center(self):
        return self._center

    @property
    def height(self):  # heigth above earth center
        """
        :return: height of the satelite
        """
        return self._heigth

    @property
    def sun_position(self):
        """
        :return: sun position as SkyCoord object in sat frame
        """

        return self._sun_position

    @property
    def sun_position_icrs(self):
        """
        :return: sun position as SkyCood object in icrs frame
        """
        return self._sun_position.icrs

    @property
    def moon_position_icrs(self):
        """                                                                                                                                                                                                                                                                   
        :return: moon position as SkyCood object in icrs frame                                                                                                                                                                                                                
        """
        tmp_moon = get_moon(self._time)
        # in ICRS
        return SkyCoord(
            tmp_moon.ra.deg,
            tmp_moon.dec.deg,
            unit="deg",
            frame="gcrs",
            obstime=self._time,
        ).icrs

    @property
    def sun_angle(self):
        """
        :return: angle between det pointing and sun position
        """

        return self._center.separation(self._sun_position)

    @property
    def sun_earth_angle(self):
        """
        :return: angle between center of earth an sun postition (seen from the satelite)
        """
        return self._sun_position.separation(self._earth_position)

    @property
    def earth_position(self):
        """
        :return: earth position as SkyCoord in sat frame
        """

        return self._earth_position

    @property
    def earth_position_icrs(self):
        """
        :return: earth position as SkyCoord in ICRS frame
        """
        return self._earth_position.icrs

    @property
    def earth_az_zen_sat(self):
        """
        :return: the longitude and latitude of the earth center in sat. frame
        """
        return [self._earth_position.lon.deg, self._earth_position.lat.deg]

    @property
    def det_ra_dec_icrs(self):
        """
        :return: the det pointing in icrs frame (ra, dec
        """
        return [(self._center.icrs).ra.deg, (self._center.icrs).dec.deg]

    def geo_to_gbm(self, pos_geo):
        """ Compute the transformation from heliocentric Sgr coordinates to
            spherical Galactic.
        """
        q1, q2, q3, q4 = self._quaternion  # q1,q2,q3,q4
        sc_matrix = np.zeros((3, 3))

        sc_matrix[0, 0] = q1 ** 2 - q2 ** 2 - q3 ** 2 + q4 ** 2
        sc_matrix[0, 1] = 2.0 * (q1 * q2 + q4 * q3)
        sc_matrix[0, 2] = 2.0 * (q1 * q3 - q4 * q2)
        sc_matrix[1, 0] = 2.0 * (q1 * q2 - q4 * q3)
        sc_matrix[1, 1] = -(q1 ** 2) + q2 ** 2 - q3 ** 2 + q4 ** 2
        sc_matrix[1, 2] = 2.0 * (q2 * q3 + q4 * q1)
        sc_matrix[2, 0] = 2.0 * (q1 * q3 + q4 * q2)
        sc_matrix[2, 1] = 2.0 * (q2 * q3 - q4 * q1)
        sc_matrix[2, 2] = -(q1 ** 2) - q2 ** 2 + q3 ** 2 + q4 ** 2

        X0 = np.dot(sc_matrix[0, :], pos_geo)
        X1 = np.dot(sc_matrix[1, :], pos_geo)
        X2 = np.clip(np.dot(sc_matrix[2, :], pos_geo), -1.0, 1.0)
        pos_sat = [X0, X1, X2]
        return np.array(pos_sat)

    def gbm_to_geo(self, pos_gbm):
        """ Compute the transformation from heliocentric Sgr coordinates to
            spherical Galactic.
        """
        q1, q2, q3, q4 = self._quaternion  # q1,q2,q3,q4
        sc_matrix = np.zeros((3, 3))

        sc_matrix[0, 0] = q1 ** 2 - q2 ** 2 - q3 ** 2 + q4 ** 2
        sc_matrix[0, 1] = 2.0 * (q1 * q2 + q4 * q3)
        sc_matrix[0, 2] = 2.0 * (q1 * q3 - q4 * q2)
        sc_matrix[1, 0] = 2.0 * (q1 * q2 - q4 * q3)
        sc_matrix[1, 1] = -(q1 ** 2) + q2 ** 2 - q3 ** 2 + q4 ** 2
        sc_matrix[1, 2] = 2.0 * (q2 * q3 + q4 * q1)
        sc_matrix[2, 0] = 2.0 * (q1 * q3 + q4 * q2)
        sc_matrix[2, 1] = 2.0 * (q2 * q3 - q4 * q1)
        sc_matrix[2, 2] = -(q1 ** 2) - q2 ** 2 + q3 ** 2 + q4 ** 2

        X0 = np.dot(sc_matrix[:, 0], pos_gbm)
        X1 = np.dot(sc_matrix[:, 1], pos_gbm)
        X2 = np.clip(np.dot(sc_matrix[:, 2], pos_gbm), -1.0, 1.0)
        pos_geo = [X0, X1, X2]
        return np.array(pos_geo)

    @property
    def earth_angle(self):
        """
        :return: angle between det pointing and earth position
        """
        return self._center.separation(self._earth_position)

    @property
    def sun_pos(self):
        """
        :return: sun position in sat frame expressed with x,y and z coordinate
        """
        lon, lat = self._sun_position.lon.deg, self._sun_position.lat.deg
        x = np.cos(lon) * np.cos(lat)
        y = np.sin(lon) * np.cos(lat)
        z = np.sin(lat)
        return np.array([x, y, z])

    @property
    def sun_lon_lat(self):
        """                                                                                                                                                                                                                                                                   
        :return: sun position in sat frame expressed with lon, lat                                                                                                                                                                                                 
        """
        return self._sun_position.lon.deg, self._sun_position.lat.deg

    @property
    def earth_pos(self):
        """                                                                                                                                                                                                                                                                   
        :return: earth position in sat frame expressed with x,y and z coordinate                                                                                                                                                                                             
        """
        lon, lat = self._earth_position.lon.deg, self._earth_position.lat.deg
        x = np.cos(lon) * np.cos(lat)
        y = np.sin(lon) * np.cos(lat)
        z = np.sin(lat)
        return np.array([x, y, z])

    @property
    def center(self):

        return self._center

    @property
    def name(self):

        return self._name

    @property
    def az(self):

        return self._az

    @property
    def zen(self):

        return self._zen

    @property
    def xyz(self):

        return self._xyz

    @property
    def mount_point(self):

        return self._mount_point


class NaI0(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 45.89
        self._zen = 90 - 20.58

        self._mount_point = np.array([96.1, 80.4, 107.6])

        super(NaI0, self).__init__("n0", quaternion, sc_pos, time)


class NaI1(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 45.11
        self._zen = 90 - 45.31
        self._mount_point = np.array([101.1, 72.8, 72.1])

        super(NaI1, self).__init__("n1", quaternion, sc_pos, time)


class NaI2(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 58.44
        self._zen = 90 - 90.21
        self._mount_point = np.array([109.0, 58.1, 99.0])

        super(NaI2, self).__init__("n2", quaternion, sc_pos, time)


class NaI3(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 314.87
        self._zen = 90 - 45.24
        self._mount_point = np.array([97.7, -76.3, 102.5])

        super(NaI3, self).__init__("n3", quaternion, sc_pos, time)


class NaI4(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 303.15
        self._zen = 90.0 - 90.27
        self._mount_point = np.array([109.0, -57.5, 83.6])

        super(NaI4, self).__init__("n4", quaternion, sc_pos, time)


class NaI5(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 3.35
        self._zen = 90 - 89.97
        self._mount_point = np.array([99.6, -49.7, 100.1])

        super(NaI5, self).__init__("n5", quaternion, sc_pos, time)


class NaI6(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 224.93
        self._zen = 90 - 20.43
        self._mount_point = np.array([-95.8, -80.3, 107.1])

        super(NaI6, self).__init__("n6", quaternion, sc_pos, time)


class NaI7(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 224.62
        self._zen = 90 - 46.18
        self._mount_point = np.array([-100.6, -72.5, 71.6])

        super(NaI7, self).__init__("n7", quaternion, sc_pos, time)


class NaI8(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 236.61
        self._zen = 90 - 89.97
        self._mount_point = np.array([-108.4, -57.2, 99.0])

        super(NaI8, self).__init__("n8", quaternion, sc_pos, time)


class NaI9(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 135.19
        self._zen = 90 - 45.55
        self._mount_point = np.array([-97.5, 76.5, 102.5])

        super(NaI9, self).__init__("n9", quaternion, sc_pos, time)


class NaIA(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """

        self._az = 123.73
        self._zen = 90 - 90.42
        self._mount_point = np.array([-108.7, 57.7, 83.7])

        super(NaIA, self).__init__("na", quaternion, sc_pos, time)


class NaIB(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 183.74
        self._zen = 90 - 90.32
        self._mount_point = np.array([-99.3, 50.0, 100.2])

        super(NaIB, self).__init__("nb", quaternion, sc_pos, time)


class BGO0(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 0.0
        self._zen = 0.0
        self._mount_point = np.array([126.05, 0.13, 63.32])

        super(BGO0, self).__init__("b0", quaternion, sc_pos, time)


class BGO1(GBMDetector):
    def __init__(self, quaternion, sc_pos=None, time=None):
        """

        Parameters
        ----------
        quaternion
        """
        self._az = 180.0
        self._zen = 0.0
        self._mount_point = np.array([-126.14, 0.01, 67.22])

        super(BGO1, self).__init__("b1", quaternion, sc_pos, time)



import collections

import healpy as hp
import matplotlib.pyplot as plt
import numpy as np
from astropy.coordinates import SkyCoord
from sympy import Point3D

from gbmgeometry.gbm import GBM
from gbmgeometry.utils.array_to_cmap import array_to_cmap
from gbmgeometry.spacecraft.geometry import Ray
from gbmgeometry.spacecraft.lat import LAT, LATRadiatorMinus, LATRadiatorPlus
from gbmgeometry.spacecraft.solar_panels import SolarPanelMinus, SolarPanelPlus

from gbmgeometry.spacecraft.gbm_detectors import add_rotated_cylinder

class Fermi(object):
    def __init__(self, quaternion, sc_pos=None):

        """
        
        A model of Fermi that can be plotted in 3D and read healpix maps.
        

        :param quaternion: the quaternion array 
        :param sc_pos: the spacecraft position
        """

        # build fermi by creating all the components

        self._lat = LAT()

        self._lat_radiator_plus = LATRadiatorPlus()
        self._lat_radiator_minus = LATRadiatorMinus()

        self._solar_panel_plus = SolarPanelPlus()
        self._solar_panel_minus = SolarPanelMinus()

        # build a GBM

        self._gbm = GBM(quaternion, sc_pos)

        # grab the frame

        self._frame = self._gbm.n0.center.frame

        # attach the components to fermi

        self._spacecraft_components = collections.OrderedDict()

        self._spacecraft_components[self._lat.name] = self._lat

        self._spacecraft_components[
            self._lat_radiator_minus.name
        ] = self._lat_radiator_minus
        self._spacecraft_components[
            self._lat_radiator_plus.name
        ] = self._lat_radiator_plus

        self._spacecraft_components[
            self._solar_panel_plus.name
        ] = self._solar_panel_plus
        self._spacecraft_components[
            self._solar_panel_minus.name
        ] = self._solar_panel_minus

        # add lists for each detector to rays

        self._rays = collections.OrderedDict()

        for name in self._gbm.detectors.keys():
            self._rays[name] = []

        self._intersection_points = None

    @property
    def spacecraft_components(self):

        return self._spacecraft_components

    @property
    def rays(self):

        return self._rays

    def add_ray(self, ray_coordinate, probability=None, color="#29FC5C"):

        """

        :param ray_coordinate: 
        :param probability: 
        :param color: 
        """
        for name, det in self._gbm.detectors.items():
            ray = Ray(det, ray_coordinate, probability=probability, color=color)

            self._rays[name].append(ray)

    def compute_intersections(self, *detectors):

        self._intersection_points = collections.OrderedDict()

        all_intersections = collections.OrderedDict()

        # go thru all detectors

        if len(detectors) == 0:
            dets = self._rays.keys()

        for det_name, det in self._rays.items():

            self._intersection_points[det_name] = []

            ray_dict = collections.OrderedDict()
            if det_name in dets:

                # now go through all rays

                for i, ray in enumerate(det):

                    # now all components

                    collision_info = collections.OrderedDict()

                    collision_info["surface"] = []
                    collision_info["point"] = []
                    collision_info["distance"] = []

                    for name, component in self._spacecraft_components.items():

                        # intersect the volume with the rays

                        component.intersect_ray(ray)

                        plane, point, distance = component.intersection

                        current_ray_origin = Point3D(ray.ray_origin)

                        if plane is not None and current_ray_origin.distance(
                            Point3D(point)
                        ) <= current_ray_origin.distance(Point3D(ray.detector_origin)):
                            collision_info["surface"].append("%s %s" % (name, plane))
                            collision_info["point"].append(point)
                            collision_info["distance"].append(distance)

                            self._intersection_points[det_name].append(point)

                    ray_dict[i] = collision_info

                all_intersections[det_name] = ray_dict

        return all_intersections

    def plot_fermi(
            self, ax=None, detectors=None, with_rays=False, with_intersections=False,
            plot_det_label=False, color_dets_different=False
    ):

        """

        :param ax: 
        :param detectors: 
        :param with_rays: 
        :param with_intersections: 
        :return: 
        """
        if ax is None:

            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")

        else:

            fig = ax.get_figure()

        if detectors is None:

            detectors = self._gbm.detectors.keys()

        else:

            for det in detectors:
                assert det in self._gbm.detectors.keys(), "invalid detector"

        for name, component in self._spacecraft_components.items():
            component.plot(ax)

        for name, det in self._gbm.detectors.items():

            if name in detectors:

                add_rotated_cylinder(ax, theta=np.pi/2-np.deg2rad(det.zen), phi=np.deg2rad(det.az),
                                     x_center=det.mount_point[0], y_center=det.mount_point[1],
                                     z_center=det.mount_point[2])
                
                if plot_det_label:
                    ax.text3D(det.mount_point[0]-10, det.mount_point[1]-20, det.mount_point[2]+20,
                              det.name, color='black', fontweight='bold', size=30)
                if color_dets_different:

                    colors = {"n0":"#FF4848","n1":"#D4AE00","n2":"#DD69FE","n3":"#FF8A1F",
                              "n4":"#9669FE","n5":"#B05F3C","n6":"#9219F1","n7":"#F70000",
                              "n8":"#3923D6","n9":"#A91374","na":"#62A9FF","nb":"#4A9586",
                              "b0":"#02EF72","b1":"#59955C"}
                    
                    add_rotated_cylinder(ax, theta=np.pi/2-np.deg2rad(det.zen), phi=np.deg2rad(det.az),
                                     x_center=det.mount_point[0], y_center=det.mount_point[1],
                                         z_center=det.mount_point[2], color=colors[det.name],
                                         label=det.name, alpha=0.9)
                    ax.legend()
                    

        if with_rays:

            # for all the detectors plot the rays

            for name, det in self._rays.items():

                if name in detectors:

                    for ray in det:
                        ray.plot(ax)

        if with_intersections:

            # if there are intersections
            # then plot the intersection points

            if self._intersection_points is not None:

                for name, points in self._intersection_points.items():
                    if name in detectors:
                        for point in points:
                            ax.scatter(*point, c="r")

        ax.set_xlabel("SCX")
        ax.set_ylabel("SCY")
        ax.set_zlabel("SCZ")

        ax.set_xlim(-(158.6+106)/2,(158.6+106)/2)
        ax.set_ylim(-(158.6+106)/2,(158.6+106)/2)
        ax.set_zlim(0,158.6+106)
        
        ax.grid(False)
        ax.xaxis.pane.set_edgecolor("black")
        ax.yaxis.pane.set_edgecolor("black")

        ax.view_init(10, 15)
        ax.axis('off')
        
        #ax.xaxis.pane.fill = False
        #ax.yaxis.pane.fill = False
        #ax.zaxis.pane.fill = False

        return fig

    def read_healpix_map(self, healpix_map, cmap="viridis"):

        # collect the nside of the map

        nside = hp.get_nside(healpix_map)

        # get the colors of the rays based of their values

        _, colors = array_to_cmap(healpix_map, cmap=cmap, use_log=False)

        # now go thru all the points on the sphere

        for idx, val in enumerate(healpix_map):

            if val > 0:

                # if the probability is greater than 0
                # then we need to get the sky position

                ra, dec = Fermi._pix_to_sky(idx, nside)

                # mark the color

                color = colors[idx]

                # now make a point source

                ps = SkyCoord(ra, dec, unit="deg", frame="icrs")

                # transform into the fermi frame

                ps_fermi = ps.transform_to(frame=self._frame)

                self.add_ray(ps_fermi, color=color)

    @staticmethod
    def _pix_to_sky(idx, nside):
        """Convert the pixels corresponding to the input indexes to sky coordinates (RA, Dec)"""

        theta, phi = hp.pix2ang(nside, idx)

        ra = np.rad2deg(phi)
        dec = np.rad2deg(0.5 * np.pi - theta)

        return ra, dec



import matplotlib as mpl
import matplotlib.pyplot as plt


def array_to_cmap(values, cmap, use_log=False):
    """
    Generates a color map and color list that is normalized
    to the values in an array. Allows for adding a 3rd dimension
    onto a plot

    :param values: a list a values to map into a cmap
    :param cmap: the mpl colormap to use
    :param use_log: if the mapping should be done in log space
    """

    if use_log:

        norm = mpl.colors.LogNorm(vmin=min(values), vmax=max(values))

    else:

        norm = mpl.colors.Normalize(vmin=min(values), vmax=max(values))

    cmap = plt.cm.ScalarMappable(norm=norm, cmap=cmap)

    rgb_colors = map(cmap.to_rgba, values)

    return cmap, rgb_colors



### Functions to add small cylinders as detector crystals ###

import numpy as np
import math

def rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    :param axis: Axis vector to rotate around
    :param theta: Angle to rotate
    :return: Rotation matrix
    """
    axis = np.asarray(axis)
    axis = axis / math.sqrt(np.dot(axis, axis))
    a = math.cos(theta / 2.0)
    b, c, d = -axis * math.sin(theta / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
    
    return np.array([[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
                     [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
                     [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])

def add_rotated_cylinder(ax, height=30, radius=15,
                         x_center=0, y_center=0, z_center=0, color='magenta',
                         alpha=0.5, theta=0, phi=0, ceiling=True, floor=True, label=None):
    '''
    Returns the unit cylinder that corresponds to the curve r.
    :param ax: ax of figure to add the cylinder
    :param height: Height of cylinder
    :param radius: Radius of cylinder
    :param x_center: x-coord of center of cylinder
    :param y_center: y-coord of center of cylinder
    :param z_center: z-coord of center of cylinder
    :param color: color of cylinder
    :param alpha: alpha of cylinder
    :param theta: theta angle of cylinder orientation
    :param phi: phi angle of cylinder orientation
    :param ceiling: add a ceiling to cylinder?
    :param floor: add a floor to cylinder?
    :param label: Label for det
    :return:
    '''
    
    r = np.ones(10)
    n = 20
    # ensure that r is a column vector
    r = np.atleast_2d(r)
    r_rows,r_cols = r.shape
    
    if r_cols > r_rows:
        r = r.T

    # find points along x and y axes
    points  = np.linspace(0,2*np.pi,n+1)
    x = np.cos(points)*r*radius# +x_center
    y = np.sin(points)*r*radius# +y_center

    # find points along z axis
    rpoints = np.atleast_2d(np.linspace(0,1,len(r)))
    z = np.ones((1,n+1))*rpoints.T*height#+z_center
    
    axis1 = [0,1.,0]
    axis2 = [0,0,1.]
    xr,yr,zr = np.dot(rotation_matrix(axis2, phi),
                      np.transpose(np.dot(rotation_matrix(axis1, theta),
                                          np.transpose(np.array([x,y,z]), (1,0,2))), (1,0,2)))
    if label is not None:
        surf = ax.plot_surface(xr+x_center,yr+y_center,zr+z_center, color=color, alpha=alpha,label=label)
    else:
        surf = ax.plot_surface(xr+x_center,yr+y_center,zr+z_center, color=color, alpha=alpha)

    surf._facecolors2d=surf._facecolors3d
    surf._edgecolors2d=surf._edgecolors3d
    
    # Ceiling and floor
    if floor:
        R = np.linspace(0, radius, 10)
        h = 0#z_center
        u = np.linspace(0,  2*np.pi, 100)

        x = np.outer(R, np.cos(u))#+x_center
        y = np.outer(R, np.sin(u))#+y_center

        xr,yr,hr = np.dot(rotation_matrix(axis2, phi),
                          np.transpose(np.dot(rotation_matrix(axis1, theta),
                                              np.transpose(np.array([x,y,h*np.ones_like(x)]),
                                                           (1,0,2))), (1,0,2)))

        surf=ax.plot_surface(xr+x_center,yr+y_center,hr+z_center, color=color, alpha=alpha)
        surf._facecolors2d=surf._facecolors3d
        surf._edgecolors2d=surf._edgecolors3d
    
    
    # Celling and floor
    if ceiling:
        R = np.linspace(0, radius, 10)
        h = height#z_center+height
        u = np.linspace(0,  2*np.pi, 100)

        x = np.outer(R, np.cos(u))#+x_center
        y = np.outer(R, np.sin(u))#+y_center

        xr,yr,hr = np.dot(rotation_matrix(axis2, phi),
                          np.transpose(np.dot(rotation_matrix(axis1, theta),
                                              np.transpose(np.array([x,y,h*np.ones_like(x)]),
                                                           (1,0,2))), (1,0,2)))

        surf=ax.plot_surface(xr+x_center,yr+y_center,hr+z_center, color=color, alpha=alpha)
        surf._facecolors2d=surf._facecolors3d
        surf._edgecolors2d=surf._edgecolors3d



import pkg_resources
import os
from shutil import copyfile


def get_path_of_data_file(data_file):

    path = get_path_of_data_dir()
    
    file_path = os.path.join(path,data_file)

    return file_path


def get_path_of_data_dir():
    file_path = pkg_resources.resource_filename("gbmgeometry", "data")

    return file_path



from gbmgeometry.spacecraft.geometry import Volume


class SolarPanel(Volume):
    def __init__(self, sign, name):
        lenght = 492.3
        width = 151.0

        xtra_y_mount = 150.0

        x_origin = 0.0
        y_origin = sign * (lenght / 2.0 + xtra_y_mount)
        z_origin = 41.5

        y_width = lenght
        x_width = 1.0
        height = width

        super(SolarPanel, self).__init__(
            name=name,
            x_origin=x_origin,
            y_origin=y_origin,
            z_origin=z_origin,
            height=height,
            x_width=x_width,
            y_width=y_width,
            color="#3498DB",
            active_surfaces=["+x", "-x"],
        )


class SolarPanelPlus(SolarPanel):
    def __init__(self):
        super(SolarPanelPlus, self).__init__(1.0, "Solar Panel +")


class SolarPanelMinus(SolarPanel):
    def __init__(self):
        super(SolarPanelMinus, self).__init__(-1.0, "Solar Panel -")



import astropy.time as time
import astropy.units as u
import numpy as np


class GBMTime(object):
    def __init__(self, time_object):

        self._time_object = time_object

        self._current_mjd = self._time_object.mjd

        # get the Fermi MET from the MET

        self._calculate_met_from_mjd()

        self._utc_zero = self._calculate_MJD_from_MET(0)

        # this is when week 9 of the mission starts
        self._utc_start_of_sc_data = "2008-08-07T03:35:44.0"
        self._time_of_start_of_sc_data = time.Time(self._utc_start_of_sc_data)

    @property
    def met(self):

        return self._met

    @property
    def utc(self):

        return self._time_object.iso

    @property
    def time(self):

        return self._time_object

    @property
    def t_zero(self):
        return self._utc_zero

    @property
    def mission_week(self):

        dt = (self._time_object - self._time_of_start_of_sc_data).to(u.week)
        return dt + 10 * u.week

    @classmethod
    def from_UTC_fits(cls, date_string):
        """

        Create a time object from a fits UTC representation


        :param date_string: 
        :return: 
        """

        time_object = time.Time(date_string, format="fits", scale="utc")

        return cls(time_object)

    @classmethod
    def from_MET(cls, met):

        time_object = GBMTime._calculate_MJD_from_MET(met)

        return cls(time_object)

    @staticmethod
    def _calculate_MJD_from_MET(met):

        if met <= 252460801.000:
            utc_tt_diff = 65.184
        elif met <= 362793602.000:
            utc_tt_diff = 66.184
        elif met <= 457401603.000:
            utc_tt_diff = 67.184
        elif met <= 504921604.000:
            utc_tt_diff = 68.184
        else:
            utc_tt_diff = 69.184

        mjdutc = (
            ((met - utc_tt_diff) / 86400.0) + 51910 + 0.0007428703703
        )  # -68.184 added to account for diff between TT and UTC and the 4 leapseconds since 2001
        # mjdtt = ((met) / 86400.0) + 51910 + 0.00074287037037

        return time.Time(mjdutc, scale="utc", format="mjd")

    def _calculate_met_from_mjd(self):
        """
        calculated the Fermi MET given MJD
        :return: 
        """

        if self._current_mjd <= 54832.00000000:
            utc_tt_diff = 65.184
        elif self._current_mjd <= 56109.00000000:
            utc_tt_diff = 66.184
        elif self._current_mjd <= 57204.00000000:
            utc_tt_diff = 67.184
        elif self._current_mjd <= 57754.00000000:
            utc_tt_diff = 68.184
        else:
            utc_tt_diff = 69.184

        self._met = (
            self._current_mjd - 51910 - 0.0007428703703
        ) * 86400.0 + utc_tt_diff  # convert it into MET

    def __add__(self, other):

        if isinstance(other, time.TimeDelta):

            new_time = self._time_object + other

        else:

            # assuming second addition

            dt = time.TimeDelta(other, format="sec")

            new_time = self._time_object + dt

        return GBMTime(new_time)

    def __sub__(self, other):

        if isinstance(other, time.TimeDelta):

            new_time = self._time_object - other

        elif isinstance(other, GBMTime):

            dt = self._time_object - other.time

            return dt

        else:

            # assuming second addition

            dt = time.TimeDelta(other, format="sec")

            new_time = self._time_object - dt

        return GBMTime(new_time)


# def mission_week(met):



from gbmgeometry.spacecraft.geometry import Volume


class LAT(Volume):
    def __init__(self):
        # cm

        height = 106.0
        z_origin = 158.6 + 106.0 / 2.0
        x_origin = 0.0
        y_origin = 0.0
        x_width = 190.3
        y_width = 190.3

        active_surfaces = ["+x", "-x", "+y", "-y", "+z"]

        super(LAT, self).__init__(
            name="LAT",
            x_origin=x_origin,
            y_origin=y_origin,
            z_origin=z_origin,
            height=height,
            x_width=x_width,
            y_width=y_width,
            color="#0000CE",
            active_surfaces=active_surfaces,
        )


class LATRadiator(Volume):
    def __init__(self, name, sign, active_surfaces):
        # cm

        height = 158.6
        z_origin = 158.6 / 2.0
        x_origin = 0.0
        y_origin = sign * 96.2
        x_width = 190.3
        y_width = 5.2

        super(LATRadiator, self).__init__(
            name=name,
            x_origin=x_origin,
            y_origin=y_origin + sign * y_width / 2,
            z_origin=z_origin,
            height=height,
            x_width=x_width,
            y_width=y_width,
            color="darkgrey",
            active_surfaces=active_surfaces,
        )


class LATRadiatorMinus(LATRadiator):
    def __init__(self):
        active_surfaces = ["+x", "-x", "+y", "-z"]

        super(LATRadiatorMinus, self).__init__("LAT Radiator-", -1, active_surfaces)


class LATRadiatorPlus(LATRadiator):
    def __init__(self):
        active_surfaces = ["+x", "-x", "-y", "-z"]

        super(LATRadiatorPlus, self).__init__("LAT Radiator+", 1, active_surfaces)



import collections

import numpy as np
from matplotlib import colors
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from mpl_toolkits.mplot3d import Axes3D
from sympy import Plane, Point3D, Line3D


class Ray(object):
    # scaling for distance of lines
    _R = 1e10  # cm
    _scale = 3e-8

    def __init__(self, detector, point_source, color="#29FC5C", probability=None):
        self._detector = detector

        self._probability = probability

        self._color = color

        self._point_source = point_source

        self._calculate_ray_origin()

    def _calculate_ray_origin(self):
        theta = np.deg2rad(90.0 - self._point_source.lat.value)
        phi = np.deg2rad(self._point_source.lon.value)

        x = Ray._R * np.cos(phi) * np.sin(theta)
        y = Ray._R * np.sin(phi) * np.sin(theta)
        z = Ray._R * np.cos(theta)

        # this is the "distant orgin of the ray"
        self._origin = np.array([x, y, z])

        self._sympy_line = Line3D(
            Point3D(self._detector.mount_point), Point3D(self._origin)
        )

        self._plot_origin = self.point_on_ray(Ray._scale)

    def plot(self, ax):
        ax.plot(
            [self._plot_origin[0], self.detector_origin[0]],
            [self._plot_origin[1], self.detector_origin[1]],
            [self._plot_origin[2], self.detector_origin[2]],
            color=self._color,
            alpha=0.8,
        )

    @property
    def detector_name(self):
        return self._detector.name

    @property
    def probability(self):
        return self._probability

    @property
    def detector_origin(self):
        return self._detector.mount_point

    @property
    def ray_origin(self):
        return self._origin

    @property
    def sympy_line(self):
        return self._sympy_line

    def point_on_ray(self, t=0.5):
        """
        get a parametrized point on the ray
        
        :param t: point between 0 and 1
        :return: 
        """

        assert 0.0 <= t <= 1.0, "t must be between 0 and 1"

        return self.detector_origin + (self._origin - self.detector_origin) * t


class Surface(object):
    def __init__(self, name, vertices):

        self._vertices = vertices

        self._name = name

        if "+" in name:

            self._sign = 1.0

        elif "-" in name:

            self._sign = -1.0

        else:

            raise RuntimeError("the plane name is wrong")

        self._calculate_origin()

    def _calculate_origin(self):
        """
        
        compute the origin of the plane
        
        
        :return: 
        """

        if "x" in self._name:

            assert len(np.unique(self._vertices[:, 0])) == 1, "vertices are wrong!"

            self._normal = np.array([1, 0, 0]) * self._sign

            x_origin = self._vertices[0, 0]

            y_origin = (min(self._vertices[:, 1]) + max(self._vertices[:, 1])) / 2.0

            z_origin = (min(self._vertices[:, 2]) + max(self._vertices[:, 2])) / 2.0

            # self._edges = [self.]

        elif "y" in self._name:

            assert len(np.unique(self._vertices[:, 1])) == 1, "vertices are wrong!"

            self._normal = np.array([0, 1.0, 0]) * self._sign

            x_origin = (min(self._vertices[:, 0]) + max(self._vertices[:, 0])) / 2.0

            y_origin = self._vertices[0, 1]

            z_origin = (min(self._vertices[:, 2]) + max(self._vertices[:, 2])) / 2.0

        elif "z" in self._name:

            assert len(np.unique(self._vertices[:, 2])) == 1, "vertices are wrong!"

            self._normal = np.array([0, 0, 1]) * self._sign

            x_origin = (min(self._vertices[:, 0]) + max(self._vertices[:, 0])) / 2.0

            y_origin = (min(self._vertices[:, 1]) + max(self._vertices[:, 1])) / 2.0

            z_origin = self._vertices[0, 2]

        self._xmax = self._vertices[:, 0].max()
        self._ymax = self._vertices[:, 1].max()
        self._zmax = self._vertices[:, 2].max()

        self._xmin = self._vertices[:, 0].min()
        self._ymin = self._vertices[:, 1].min()
        self._zmin = self._vertices[:, 2].min()

        self._origin = np.array([x_origin, y_origin, z_origin])

        # TODO: perhaps rewrite this to find the plane

        self._sympy_plane = Plane(Point3D(self._origin), normal_vector=self._normal)

    def is_intersecting(self, ray):
        """
        checks if ray intersects plane
        
        :param ray: 
        :return: bool, array
        """

        intersecting_point = self._sympy_plane.intersection(ray.sympy_line)[0]

        if "x" in self._name:

            if self._within_y_bounds(intersecting_point.y) and self._within_z_bounds(
                intersecting_point.z
            ):
                return (
                    True,
                    np.array(
                        map(
                            float,
                            [
                                intersecting_point.x,
                                intersecting_point.y,
                                intersecting_point.z,
                            ],
                        )
                    ),
                )

        elif "y" in self._name:

            if self._within_x_bounds(intersecting_point.x) and self._within_z_bounds(
                intersecting_point.z
            ):
                return (
                    True,
                    np.array(
                        map(
                            float,
                            [
                                intersecting_point.x,
                                intersecting_point.y,
                                intersecting_point.z,
                            ],
                        )
                    ),
                )

        elif "z" in self._name:

            if self._within_y_bounds(intersecting_point.y) and self._within_x_bounds(
                intersecting_point.x
            ):
                return (
                    True,
                    np.array(
                        map(
                            float,
                            [
                                intersecting_point.x,
                                intersecting_point.y,
                                intersecting_point.z,
                            ],
                        )
                    ),
                )

        return False, None

    def _within_x_bounds(self, x):

        if x <= self._xmax and self._xmin <= x:

            return True

        else:

            return False

    def _within_y_bounds(self, y):

        if y <= self._ymax and self._ymin <= y:

            return True

        else:

            return False

    def _within_z_bounds(self, z):

        if z <= self._zmax and self._zmin <= z:

            return True

        else:

            return False

    @property
    def origin(self):

        return self._origin


class Volume(object):
    def __init__(
        self,
        name,
        x_origin,
        y_origin,
        z_origin,
        height,
        x_width,
        y_width,
        color="grey",
        active_surfaces=None,
    ):

        if active_surfaces is None:

            self._active_surfaces = ("+x", "-x", "+y", "-y", "+z", "-z")

        else:

            for surface in active_surfaces:
                assert surface in [
                    "+x",
                    "-x",
                    "+y",
                    "-y",
                    "+z",
                    "-z",
                ], "not a valid surface!"

            self._active_surfaces = active_surfaces

        self._center = (x_origin, y_origin, z_origin)

        self._color = colors.to_rgb(color)

        self._name = name

        self._build_cube(
            origin=(
                x_origin - x_width / 2.0,
                z_origin - height / 2.0,
                y_origin - y_width / 2.0,
            ),
            width=x_width,
            depth=y_width,
            height=height,
        )

        self._intersections = None

    def _build_cube(self, origin=None, width=1, height=1, depth=1):

        self._planes = collections.OrderedDict()
        for plane in ["+x", "-x", "+y", "-y", "+z", "-z"]:
            self._planes[plane] = None

        u, v, w = (0, 0, 0) if origin is None else origin

        grids = []

        for plane in self._planes.keys():

            if "-z" in plane:
                this_grid = self._grid("xy", (u, w), width, depth, v)

                self._planes[plane] = Surface(plane, this_grid[0])

                grids.extend(this_grid)

            if "+z" in plane:
                this_grid = self._grid("xy", (u, w), width, depth, v + height)

                self._planes[plane] = Surface(plane, this_grid[0])

                grids.extend(this_grid)

            if "-y" in plane:
                this_grid = self._grid("xz", (u, v), width, height, w)

                self._planes[plane] = Surface(plane, this_grid[0])

                grids.extend(this_grid)

            if "+y" in plane:
                this_grid = self._grid("xz", (u, v), width, height, w + depth)

                self._planes[plane] = Surface(plane, this_grid[0])

                grids.extend(this_grid)

            if "-x" in plane:
                this_grid = self._grid("yz", (w, v), depth, height, u)

                self._planes[plane] = Surface(plane, this_grid[0])

                grids.extend(this_grid)

            if "+x" in plane:
                this_grid = self._grid("yz", (w, v), depth, height, u + width)

                self._planes[plane] = Surface(plane, this_grid[0])

                grids.extend(this_grid)

        self._quads = np.array(grids)

    @staticmethod
    def _grid(plane="xy", origin=None, width=1, height=1, depth=0):

        u, v = (0, 0) if origin is None else origin

        # hard code this
        width_segments = 1
        height_segments = 1

        w_x, h_y = width, height

        quads = []

        for i in range(width_segments):
            for j in range(height_segments):
                quads.append(
                    Volume._quad(plane, (i * w_x + u, j * h_y + v), w_x, h_y, depth)
                )

        return np.array(quads)

    @staticmethod
    def _quad(plane="xy", origin=None, width=1, height=1, depth=0):
        u, v = (0, 0) if origin is None else origin

        plane = plane.lower()
        if plane == "xy":
            vertices = (
                (u, v, depth),
                (u + width, v, depth),
                (u + width, v + height, depth),
                (u, v + height, depth),
            )
        elif plane == "xz":
            vertices = (
                (u, depth, v),
                (u + width, depth, v),
                (u + width, depth, v + height),
                (u, depth, v + height),
            )
        elif plane == "yz":
            vertices = (
                (depth, u, v),
                (depth, u + width, v),
                (depth, u + width, v + height),
                (depth, u, v + height),
            )
        else:
            raise ValueError('"{0}" is not a supported plane!'.format(plane))

        return np.array(vertices)

    @property
    def planes(self):
        return self._planes

    @property
    def center(self):
        return self._center

    @property
    def name(self):

        return self._name

    def plot(self, ax, alpha=0.1):

        collection = Poly3DCollection(self._quads, facecolors=self._color, alpha=0.25)

        c = []

        for i in self._color:
            c.append(i)

        c.append(alpha)

        collection.set_facecolor(c)

        ax.add_collection3d(collection)

    def intersect_ray(self, ray):

        intersections = collections.OrderedDict()

        for k, v in self._planes.iteritems():

            if k in self._active_surfaces:

                intersection_info = collections.OrderedDict()

                is_intersecting, point = v.is_intersecting(ray)

                if is_intersecting:
                    intersection_info["intersection point"] = point

                    # now get the distance between the points

                    d2 = (np.power(point - ray.detector_origin, 2)).sum()

                    intersection_info["distance"] = np.sqrt(d2)

                    intersections[k] = intersection_info

        self._intersections = intersections

    @property
    def all_intersections(self):

        return self._intersections

    @property
    def intersection(self):

        # return the first intersection

        max_distance = 0.0

        intersection = None

        for k, v in self._intersections.iteritems():

            if v["distance"] > max_distance:
                intersection = k
                max_distance = v["distance"]

        if intersection is None:
            return None, None, None

        return (
            intersection,
            self._intersections[intersection]["intersection point"],
            self._intersections[intersection]["distance"],
        )



import astropy.io.fits as fits
import astropy.units as u
import numpy as np
import scipy.interpolate as interpolate

from gbmgeometry.utils.gbm_time import GBMTime


class PositionInterpolator(object):
    def __init__(self, poshist=None, T0=None, trigdat=None):

        # Use position history file
        """

        Parameters
        ----------
        poshist
        T0
        trigdat

        Returns
        -------

        """
        if poshist is not None:

            with fits.open(poshist) as poshist:
                # poshist = fits.open(poshist)
                self._time = poshist["GLAST POS HIST"].data["SCLK_UTC"]
                self._quats = np.array(
                    [
                        poshist["GLAST POS HIST"].data["QSJ_1"],
                        poshist["GLAST POS HIST"].data["QSJ_2"],
                        poshist["GLAST POS HIST"].data["QSJ_3"],
                        poshist["GLAST POS HIST"].data["QSJ_4"],
                    ]
                ).T

                self._sc_pos = np.array(
                    [
                        poshist["GLAST POS HIST"].data["POS_X"],
                        poshist["GLAST POS HIST"].data["POS_Y"],
                        poshist["GLAST POS HIST"].data["POS_Z"],
                    ]
                ).T

                # if using posthist then units are in m

                self._factor = (u.m).to(u.km)

                if T0 is not None:
                    self._time -= T0

                    self._trigtime = T0

                else:

                    self._trigtime = None

            # poshist.close()

        elif trigdat is not None:

            with fits.open(trigdat) as trigdat:
                # trigdat = fits.open(trigdat)
                trigtime = trigdat["EVNTRATE"].header["TRIGTIME"]
                tstart = trigdat["EVNTRATE"].data["TIME"] - trigtime

                self._trigtime = trigtime

                self._quats = trigdat["EVNTRATE"].data["SCATTITD"]
                self._sc_pos = trigdat["EVNTRATE"].data["EIC"]

                sort_mask = np.argsort(tstart)
                tstart = tstart[sort_mask]

                self._quats = self._quats[sort_mask]
                self._sc_pos = self._sc_pos[sort_mask]

                self._time = tstart

            # trigdat.close()

            # the sc is in km so no need to convert

            self._factor = 1

        else:

            print("No file passed. Exiting")
            return

        # Interpolate the stuf
        self._interpolate_quaternion()
        self._interpolate_sc_pos()

    def utc(self, t):

        if self._trigtime is not None:

            met = self._trigtime + t

        else:

            met = t

        time = GBMTime.from_MET(met)
        # print(time.time.fits)
        return time.time.fits

    def met(self, t):

        if self._trigtime is not None:

            met = self._trigtime + t

        else:

            met = t

        return met

    def maxtime(self):

        return self._time

    def quaternion(self, t):
        """
        Gets an interpolated quaternion as a function of time

        Parameters
        ----------
        t

        Returns
        -------
        A Fermi GBM quaternion

        """

        return self._quaternion_t(t)

    def sc_pos(self, t):
        """

        Parameters
        ----------
        t

        Returns
        -------
        Fermi GBM spacecraft position

        """
        return self._scxyz_t(t) * self._factor

    def _interpolate_quaternion(self):

        self._quaternion_t = interpolate.interp1d(self._time, self._quats.T)

    def _interpolate_sc_pos(self):

        self._scxyz_t = interpolate.interp1d(self._time, self._sc_pos.T)

    def sc_matrix(self, t):

        q1, q2, q3, q4 = self.quaternion(t)
        sc_matrix = np.zeros((3, 3))

        sc_matrix[0, 0] = q1 ** 2 - q2 ** 2 - q3 ** 2 + q4 ** 2
        sc_matrix[0, 1] = 2.0 * (q1 * q2 + q4 * q3)
        sc_matrix[0, 2] = 2.0 * (q1 * q3 - q4 * q2)
        sc_matrix[1, 0] = 2.0 * (q1 * q2 - q4 * q3)
        sc_matrix[1, 1] = -(q1 ** 2) + q2 ** 2 - q3 ** 2 + q4 ** 2
        sc_matrix[1, 2] = 2.0 * (q2 * q3 + q4 * q1)
        sc_matrix[2, 0] = 2.0 * (q1 * q3 + q4 * q2)
        sc_matrix[2, 1] = 2.0 * (q2 * q3 - q4 * q1)
        sc_matrix[2, 2] = -(q1 ** 2) - q2 ** 2 + q3 ** 2 + q4 ** 2

        return sc_matrix

    def geo_matrix(self, t):

        return self.sc_matrix(t).T

    def altitude(self, t):
        """
        
        :param t: 
        :return: 
        """

        earth_radius = 6371.0
        fermi_radius = np.sqrt((self.sc_pos(t) ** 2).sum())

        return fermi_radius - earth_radius

    @staticmethod
    def normalize(x):
        norm = np.sqrt(np.sum(x ** 2, axis=0))
        return x / norm



from gbmgeometry.gbm import GBM#, get_legal_pairs
from gbmgeometry.gbm_detector import BGO0, BGO1
from gbmgeometry.gbm_detector import NaI0, NaI1, NaI2, NaI3, NaI4, NaI5
from gbmgeometry.gbm_detector import NaI6, NaI7, NaI8, NaI9, NaIA, NaIB
from gbmgeometry.gbm_frame import GBMFrame
from gbmgeometry.position_interpolator import PositionInterpolator
from gbmgeometry.spacecraft.fermi import Fermi


gbm_detector_list = {
    "n0": NaI0,
    "n1": NaI1,
    "n2": NaI2,
    "n3": NaI3,
    "n4": NaI4,
    "n5": NaI5,
    "n6": NaI6,
    "n7": NaI7,
    "n8": NaI8,
    "n9": NaI9,
    "na": NaIA,
    "nb": NaIB,
    "b0": BGO0,
    "b1": BGO1,
}


from gbmgeometry.utils.gbm_time import GBMTime

__all__ = [
    "GBMFrame",
    "GBM",
    "NaI0",
    "NaI1",
    "NaI2",
    "NaI3",
    "NaI4",
    "NaI5",
    "NaI6",
    "NaI7",
    "NaI8",
    "NaI9",
    "NaIA",
    "NaIB",
    "BGO0",
    "BGO1",
    "PositionInterpolator",
#    "get_legal_pairs",
    "Fermi",
    "GBMTime",
    "gbm_detector_list",
]



from setuptools import setup, find_packages, Command
import os
import io
import sys
from shutil import rmtree


# Package meta-data.
NAME = "gbmgeometry"
DESCRIPTION = "Geometry calculations for Fermi GBM"
URL = "https://github.com/mpe-grb/gbmgeometry"
EMAIL = "jmichaelburgess@gmail.com"
AUTHOR = "J. Michael Burgess"
REQUIRES_PYTHON = ">=2.7.0"
VERSION = None

REQUIRED = [
    "numpy",
    "matplotlib",
    "astropy",
    "sympy",
    "healpy",
    "spherical_geometry",
]

# What packages are optional?
EXTRAS = {
    # 'fancy feature': ['django'],
}


here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds...")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution...")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine...")
        os.system("twine upload dist/*")

        self.status("Pushing git tags...")
        os.system("git tag v{0}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


# Create list of data files
def find_data_files(directory):

    paths = []

    for (path, directories, filenames) in os.walk(directory):

        for filename in filenames:

            paths.append(os.path.join("..", path, filename))

    return paths


# extra_files = find_data_files('/data')


setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=("tests",)),
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    #    package_data={'': extra_files, },
    license="BSD",
    cmdclass={"upload": UploadCommand,},
)
